import abc
from . import interval_utils


class _NoteGroup(abc.ABC):
    """
    An abstract base class used to define an interface for note groups, such as scales and chords.
    """
    @abc.abstractmethod
    def __init__(self, grouptype, qualified_name):
        """ Represents a group such as a scale or chord """
        builder = _get_group_builder(grouptype)  # get the function that builds the group
        unpacked_name = _unpack_group_name(qualified_name)
        self._validate_root(unpacked_name)
        self.root = unpacked_name['ROOT']
        self.quality = unpacked_name['QUALITY']
        try:
            self.notes = builder(self.root, self.quality)
        except KeyError:
            raise InvalidQualityError("Quality %s is not valid" % self.quality) from None

    def _validate_root(self, unpacked_name):
        """ Needed to validate the root of certain subclasses """
        pass

    def __str__(self):
        return str(self.notes)

    def __repr__(self):
        return str(self.notes)

    def __eq__(self, other):
        return self.notes == other.notes

    @abc.abstractmethod
    def __getitem__(self, item):
        pass

    def __contains__(self, note):
        note = note[0].upper() + note[1:]
        return note in self.notes


def _unpack_group_name(groupname):
    split_group = groupname.split(' ', 1)
    try:
        root = split_group[0][0].upper() + split_group[0][1:]
        quality = split_group[1].upper()
    except IndexError:
        raise InvalidQualityError("No quality given.") from None
    return {'ROOT': root, 'QUALITY': quality}


def _get_group_builder(grouptype):
    if grouptype == 'SCALE':
        return _build_scale
    elif grouptype == 'CHORD':
        return _build_chord


def _build_scale(root, quality):
    scale_intervals = {'MAJOR': [0, 2, 4, 5, 7, 9, 11],
                       'NATURAL MINOR': [0, 2, 3, 5, 7, 8, 10],
                       'HARMONIC MINOR': [0, 2, 3, 5, 7, 8, 11],
                       'MELODIC MINOR': [0, 2, 3, 5, 7, 9, 11],
                       }
    return _build_group(root, scale_intervals[quality])


def _build_chord(root, quality):
    chord_intervals = {'MAJOR': [0, 4, 7],
                       'MINOR': [0, 3, 7],
                       'DIMINISHED': [0, 3, 6],
                       'AUGMENTED': [0, 4, 108],
                       'DOMINANT 7': [0, 4, 7, 10],
                       'MAJOR 7': [0, 4, 7, 11],
                       'MINOR 7': [0, 3, 7, 10],
                       'MINOR 7b5': [0, 3, 6, 10],
                       'DOMINANT 9': [0, 4, 7, 10, 2],
                       'DOMINANT b9': [0, 4, 7, 10, 1],
                       'DOMINANT #9': [0, 4, 7, 10, 103],
                       'MAJOR 9': [0, 4, 7, 11, 2],
                       'MINOR 9': [0, 3, 7, 10, 2],
                       'DOMINANT 11': [0, 4, 7, 10, 2, 5],
                       'DOMINANT b11': [0, 4, 7, 10, 2, 104],
                       'DOMINANT #11': [0, 4, 7, 10, 2, 106],
                       'MAJOR 11': [0, 4, 7, 11, 2, 5],
                       'MINOR 11': [0, 3, 7, 10, 2, 5],
                       'DOMINANT 13': [0, 4, 7, 10, 2, 5, 9],
                       'DOMINANT b13': [0, 4, 7, 10, 2, 5, 8],
                       'DOMINANT #13': [0, 4, 7, 10, 2, 5, 110],
                       'MAJOR 13': [0, 4, 7, 11, 2, 5, 9],
                       'MINOR 13': [0, 3, 7, 10, 2, 5, 9]
                       }
    return _build_group(root, chord_intervals[quality])


def _build_group(root, intervals):
    # build a group of notes (scale, triad, chord) from root using the list of intervals.
    return tuple([interval_utils.INTERVAL_NOTE_PAIRS[root][interval] for interval in intervals])


class InvalidDegreeError(Exception):
    """
    Raised when an attempting to fetch an invalid scale degree name. Valid scale degree names are
    tonic, supertonic, mediant, subdominant, dominant, submediant, and leading tone.
    """
    pass


class InvalidQualityError(Exception):
    """
    Raised when the quality of a scale or chord is invalid.
    """
    pass



from .notegroups import _NoteGroup, InvalidDegreeError
from .chords import Chord


class Scale(_NoteGroup):
    """
    Represents a collection of notes. Scales are built from a series of whole and half steps and have a key signature
    and tonic. Each note in a scale is identified either by a number (1 through 7) or a degree name. Valid tonics are
    English letters A through G, and valid qualities are MAJOR, HARMONIC MINOR, MELODIC MINOR, and NATURAL MINOR. An
    InvalidTonicError is raised if the scale name has a key signature involving double sharps or double flats.
    """
    _VALID_SCALE_NAMES = {'MAJOR': ['A', 'B', 'C', 'D', 'E', 'F', 'G',
                                    'C#', 'F#',
                                    'Ab', 'Bb', 'Cb', 'Db', 'Eb', 'Gb'],
                          'MINOR': ['A', 'B', 'C', 'D', 'E', 'F', 'G',
                                    'A#', 'D#', 'G#', 'C#', 'F#',
                                    'Ab', 'Eb', 'Bb']}

    _SHARPS = ['F#', 'C#', 'G#', 'D#', 'A#', 'E#', 'B#']
    _FLATS = ['Bb', 'Eb', 'Ab', 'Db', 'Gb', 'Cb', 'Fb']

    _KEY_SIGNATURE_NUMBERS = {'C MAJOR': 0, 'G MAJOR': 1, 'D MAJOR': 2, 'A MAJOR': 3, 'E MAJOR': 4, 'B MAJOR': 5,
                              'F# MAJOR': 6, 'C# MAJOR': 7,
                              'F MAJOR': -1, 'Bb MAJOR': -2, 'Eb MAJOR': -3, 'Ab MAJOR': -4, 'Db MAJOR': -5,
                              'Gb MAJOR': -6, 'Cb MAJOR': -7,
                              'A MINOR': 0, 'E MINOR': 1, 'B MINOR': 2, 'F# MINOR': 3, 'C# MINOR': 4, 'G# MINOR': 5,
                              'D# MINOR': 6, 'A# MINOR': 7,
                              'D MINOR': -1, 'G MINOR': -2, 'C MINOR': -3, 'F MINOR': -4, 'Bb MINOR': -5,
                              'Eb MINOR': -6, 'Ab MINOR': -7}

    _KEY_SIGNATURES = {0: [], 1: _SHARPS[:1], 2: _SHARPS[:2], 3: _SHARPS[:3], 4: _SHARPS[:4], 5: _SHARPS[:5],
                       6: _SHARPS[:6], 7: _SHARPS[:7],
                       -1: _FLATS[:1], -2: _FLATS[:2], -3: _FLATS[:3], -4: _FLATS[:4], -5: _FLATS[:5],
                       -6: _FLATS[:6], -7: _FLATS[:7]}

    def __init__(self, qualified_name):
        super().__init__('SCALE', qualified_name)
        self.key_signature = Scale._fetch_key_signature(self.root, self.quality)

    def __getitem__(self, degree):
        """
        :param degree: a string representing the scale degree, such as TONIC, MEDIANT.
        :return: a Note object representing the scale degree.
        """
        degree_names = {'TONIC': 0, 'SUPERTONIC': 1, 'MEDIANT': 2, 'SUBDOMINANT': 3,
                        'DOMINANT': 4, 'SUBMEDIANT': 5, 'SEVENTH': 6, 'LEADING TONE': 6}

        try:
            degree = degree.upper()
            degree_number = degree_names[degree]
            return self.notes[degree_number]
        except KeyError:
            raise InvalidDegreeError('Invalid degree name: %s' % degree) from None

    def get_relative(self):
        """
        Returns the relative major or minor of the current scale.

        :rtype: Scale
        """
        if self.quality == 'MAJOR':
            relative_quality = 'NATURAL MINOR'
            relative_tonic = self.__getitem__('SUBMEDIANT')
        else:
            relative_quality = 'MAJOR'
            relative_tonic = self.__getitem__('MEDIANT')
        qualified_relative_name = '%s %s' % (relative_tonic, relative_quality)
        return Scale(qualified_relative_name)

    def get_parallel(self):
        """
        Returns the parallel major or minor of the current scale.

        :rtype: Scale
        """
        if self.quality == 'MAJOR':
            parallel_quality = 'NATURAL MINOR'
        else:
            parallel_quality = 'MAJOR'
        qualified_parallel_name = '%s %s' % (self.root, parallel_quality)
        return Scale(qualified_parallel_name)

    def get_triad_for_degree(self, degree):
        """
        Builds a triad based on the current scale's quality and the given degree.

        :param degree: a string representing the scale degree, such as TONIC, MEDIANT.
        :raises: InvalidDegreeError
        :return: a Chord object with the scale degree as its root.
        """
        degree = degree.upper()
        if self.quality == 'MAJOR':
            triad_qualities = {'TONIC': 'MAJOR', 'SUPERTONIC': 'MINOR', 'MEDIANT': 'MINOR', 'SUBDOMINANT': 'MAJOR',
                               'DOMINANT': 'MAJOR', 'SUBMEDIANT': 'MINOR', 'SEVENTH': 'DIMINISHED', 'LEADING TONE': 'DIMINISHED'}
        elif self.quality == 'NATURAL MINOR':
            triad_qualities = {'TONIC': 'MINOR', 'SUPERTONIC': 'DIMINISHED', 'MEDIANT': 'MAJOR', 'SUBDOMINANT': 'MINOR',
                               'DOMINANT': 'MINOR', 'SUBMEDIANT': 'MAJOR', 'SEVENTH': 'MAJOR', 'LEADING TONE': 'MAJOR'}
        elif self.quality == 'HARMONIC MINOR':
            triad_qualities = {'TONIC': 'MINOR', 'SUPERTONIC': 'DIMINISHED', 'MEDIANT': 'AUGMENTED',
                               'SUBDOMINANT': 'MINOR', 'DOMINANT': 'MAJOR', 'SUBMEDIANT': 'MAJOR', 'SEVENTH': 'DIMINISHED',
                               'LEADING TONE': 'DIMINISHED'}
        else:
            triad_qualities = {'TONIC': 'MINOR', 'SUPERTONIC': 'MINOR', 'MEDIANT': 'AUGMENTED', 'SUBDOMINANT': 'MAJOR',
                               'DOMINANT': 'MAJOR', 'SUBMEDIANT': 'DIMINISHED', 'SEVENTH': 'DIMINISHED',
                               'LEADING TONE': 'DIMINISHED'}

        triad_root = self.__getitem__(degree)  # Raises InvalidDegreeError
        triad_quality = triad_qualities[degree]
        triad_name = "%s %s" % (triad_root, triad_quality)
        return Chord(triad_name)

    def _validate_root(self, unpacked_name):
        """
        Ensures that the tonic of the scale is valid.

        :param list unpacked_name: The qualified name of the scale unpacked into a list.
        :raises: InvalidTonicError: If the tonic is an invalid note or if the key signature of the
            scale would contain double sharps or double flats.
        """
        if 'MINOR' in unpacked_name['QUALITY']:
            valid_tonics = Scale._VALID_SCALE_NAMES['MINOR']
        else:
            valid_tonics = Scale._VALID_SCALE_NAMES['MAJOR']

        if unpacked_name['ROOT'] not in valid_tonics:
            raise InvalidTonicError(
                "Invalid tonic: %s. It is possible that this tonic is a valid note name but that "
                "building the desired scale from this note would result in a scale with an invalid "
                "key signature." % unpacked_name['ROOT'])

    @classmethod
    def _fetch_key_signature(cls, tonic, quality):
        """ Returns a list of strings with the note names that make up the scale's key signature. """
        qualified_tonic = tonic + (' MINOR' if 'MINOR' in quality else ' MAJOR')
        key_signature_number = cls._KEY_SIGNATURE_NUMBERS[qualified_tonic]
        return tuple(cls._KEY_SIGNATURES[key_signature_number])


class InvalidTonicError(Exception):
    """
    Raised when attempting to create a Scale object with an invalid tonic. This situation could arise from attempting to
    create a Scale object with a tonic that is greater than G, or when attempting to create a Scale object with a valid
    tonic name (that is an English letter between A and G), but for whom the passed quality would result in a scale that
    includes invalid note names in its key signature, such as double sharps. An example of this situation is a G# Major
    scale.
    """
    pass



from .notegroups import _NoteGroup, InvalidDegreeError
from .notes import VALID_QUALIFIED_NAMES


class Chord(_NoteGroup):
    """
    Represents triads and chords. Triads have a quality of either major or minor, and consist of a bass note,
    a third, and a fifth. Chords are triads with upper extensions.
    Possible upper extensions are 7, 9, 11, and 13, optionally qualified with a sharp (#) or flat (b).
    """
    def __init__(self, qualified_name):
        """
        MAJOR, MINOR, DIMINISHED, and AUGMENTED qualities are triads. Anything with a
        number (e.g. MAJOR 7) produces a four-or-more-note chord.
        """
        super().__init__('CHORD', qualified_name)

    def __getitem__(self, element):
        degree_names = {'BASS': 0, 'THIRD': 1, 'FIFTH': 2, 'SEVENTH': 3, 'NINTH': 4, 'ELEVENTH': 5, 'THIRTEENTH': 6}
        try:
            element = element.upper()
            chord_degree = degree_names[element]
            return self.notes[chord_degree]
        except IndexError:
            return None
        except KeyError:
            raise InvalidDegreeError("Invalid degree name: %s" % element) from None

    def _validate_root(self, unpacked_name):
        """
        Ensures that the bass of the chord is valid.

        :param dict unpacked_name: The qualified name of the chord unpacked into a list.
        :raises: InvalidBassError: If the bass note is an invalid note.
        """
        if unpacked_name['ROOT'] not in VALID_QUALIFIED_NAMES:
            raise InvalidBassError("Invalid bass note: %s" % unpacked_name['ROOT'])


class InvalidBassError(Exception):
    """
    Raised when attempting to create a Chord object with an invalid bass note.
    """
    pass



from unittest import TestCase

from .. import chords
from ..notegroups import InvalidDegreeError, InvalidQualityError


class TestChords(TestCase):
    def setUp(self):
        self.c_thirteen = chords.Chord('C DOMINANT 13')
        self.c_major = chords.Chord('c major')
        self.d_minor = chords.Chord('d minor')

    def test_invalid_bass(self):
        with self.assertRaises(chords.InvalidBassError):
            z = chords.Chord('Z MAJOR')

    def test_invalid_quality(self):
        with self.assertRaises(InvalidQualityError):
            a = chords.Chord('A Foo')

    def test_getitem_bass(self):
        c = self.c_thirteen['bass']
        self.assertEqual(c, 'C')

    def test_getitem_third(self):
        e = self.c_thirteen['THIRD']
        self.assertEqual(e, 'E')

    def test_getitem_fifth(self):
        g = self.c_thirteen['FIFTH']
        self.assertEqual(g, 'G')

    def test_getitem_seventh(self):
        b_flat = self.c_thirteen['SEVENTH']
        self.assertEqual(b_flat, 'Bb')

    def test_getitem_ninth(self):
        d = self.c_thirteen['NINTH']
        self.assertEqual(d, 'D')

    def test_getitem_eleventh(self):
        f = self.c_thirteen['ELEVENTH']
        self.assertEqual(f, 'F')

    def test_getitem_thirteenth(self):
        a = self.c_thirteen['THIRTEENTH']
        self.assertEqual(a, 'A')

    def test_getitem_nonexistant_degree(self):
        b_flat = self.c_major['SEVENTH']
        self.assertTrue(b_flat is None)

    def test_getitem_invalid_degree(self):
        with self.assertRaises(InvalidDegreeError):
            bad_degree = self.c_major['baddeg']



from .interval_utils import _IntervalBuilder

VALID_NOTES = ['A', 'B', 'C', 'D', 'E', 'F', 'G']
VALID_QUALIFIERS = ['', '#', '##', 'b', 'bb']
VALID_QUALIFIED_NAMES = [name + qualifier for name in VALID_NOTES for qualifier in VALID_QUALIFIERS]


class Note:
    """
    Represents an individual note, the atomic element of Western music theory. The note's qualified name is the note
    name, optionally followed by the qualifier. Valid note names are English capital letters A through G. Valid
    qualifiers are # (sharp), b (flat), ## (double sharp), bb (double flat). Note that flats must be lowercase or a
    NoteNameError is raised. A note name followed by no qualifier represents a natural, e.g. 'A' represents 'A
    natural'. If an invalid qualified note name is passed, a NoteNameError is raised. Once a Note object is created,
    it's qualified name should not be modified.
    """
    def __init__(self, qualified_name):
        self.qualified_name = qualified_name
        self._interval_builder = _IntervalBuilder(self.qualified_name)

    @property
    def qualified_name(self):
        return self._qualified_name

    @qualified_name.setter
    def qualified_name(self, name):
        name = str(name)[0].upper() + str(name)[1:]
        if name not in VALID_QUALIFIED_NAMES:
            raise NoteNameError("Invalid note name: %s." % name)
        else:
            self._qualified_name = name

    def __str__(self):
        return self.qualified_name

    def __repr__(self):
        return self.qualified_name

    def __eq__(self, other):
        return self.qualified_name == other.qualified_name

    def ascend_interval(self, qualified_interval_name):
        """
        :param qualified_interval_name: The interval to ascend, e.g. major 3.
        :return: A Note object representing the top note of the interval. If the top note of the interval is not
            a valid note, such as F###, and InvalidIntervalError is raised.
        """
        try:
            return Note(self._interval_builder.ascend_interval_from_name(qualified_interval_name))
        except NoteNameError:
            raise NoteNameError("Ascending a %s from %s results in an invalid note." % (qualified_interval_name,
                                                                                        self._qualified_name)) from None

    def descend_interval(self, qualified_interval_name):
        """
        :param qualified_interval_name: the interval to fetch, e.g. major 3.
        :return: a Note object representing the bottom note of the interval. If the bottom note of the interval is not
            a valid note, such as F###, an InvalidInvervalError is raised.
        """
        try:
            return Note(self._interval_builder.descend_interval_from_name(qualified_interval_name))
        except NoteNameError:
            raise NoteNameError("Descending a %s from %s results in an invalid note." % (qualified_interval_name,
                                                                                         self._qualified_name)) from None

    def get_interval_name(self, top_note):
        """
        Gives the name of the interval between the current note and ``top_note``. Note that the current note is
        always treated as the bottom note in the interval.

        :param str top_note: the qualified name of the top note of the interval.
        """
        try:
            return self._interval_builder.get_interval_name(self.qualified_name, top_note)
        except NoteNameError:
            raise NoteNameError("%s is an invalid note." % top_note) from None
            

class NoteNameError(Exception):
    """
    Raised when attempting to create a Note object with a qualified note name that is not valid.
    """
    pass



import setuptools
from os import path

this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.rst'), encoding='utf-8') as readme:
    long_description = readme.read()

setuptools.setup(
    name="musictheorpy",
    version="1.3.1",
    author="Jeffrey Moorhead",
    author_email="jeff.moorhead1@gmail.com",
    description="A music theory library for Python",
    long_description=long_description,
    long_description_content_type="text/x-rst",
    url="https://github.com/Jeff-Moorhead/musictheorpy",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent"
    ],
)



import unittest
from ..notes import Note, NoteNameError
from ..interval_utils import INTERVAL_NOTE_PAIRS, _IntervalBuilder, InvalidIntervalError


class TestNote(unittest.TestCase):
    def setUp(self):
        self.a = Note('a')
        self.c = Note('C')
        self.a_flat = Note('Ab')
        self.builder = _IntervalBuilder('A')

    def test_ascend_interval(self):
        for interval, steps in self.builder.interval_steps.items():
            top_note = INTERVAL_NOTE_PAIRS['A'][steps]
            self.assertEqual(self.a.ascend_interval(interval), Note(top_note))

    def test_descend_interval(self):
        c = self.a.descend_interval("major 6")
        self.assertEqual(c, Note('C'))

    def test_invalid_qualified_name(self):
        with self.assertRaises(NoteNameError):
            z_sharp = Note('z#')

    def test_invalid_interval(self):
        with self.assertRaises(NoteNameError):
            self.a_flat.ascend_interval('DIMINISHED 2')

    def test_invalid_interval_name(self):
        with self.assertRaises(InvalidIntervalError):
            self.a_flat.ascend_interval('bad interval')

    def test_valid_top_bottom_interval(self):
        self.assertEqual(self.a.get_interval_name('C#'), "major 3")
        self.assertEqual(self.c.get_interval_name('Bb'), "minor 7")


if __name__ == '__main__':
    unittest.main()



import unittest
from ..scales import Scale, InvalidTonicError
from ..notegroups import InvalidQualityError, InvalidDegreeError
from ..chords import Chord


class TestScale(unittest.TestCase):
    def setUp(self):
        self.c_scale = Scale('c major')
        self.b_flat = Scale('Bb major')
        self.d_harm_minor = Scale('D harmonic minor')

    def test_scale(self):
        c_note_names = ('C', 'D', 'E', 'F', 'G', 'A', 'B')
        self.assertEqual(self.c_scale.root, 'C')
        self.assertEqual(self.c_scale.key_signature, ())
        self.assertEqual(self.c_scale.quality, 'MAJOR')
        self.assertEqual(self.c_scale.notes, c_note_names)

        bflat_notes = ('Bb', 'C', 'D', 'Eb', 'F', 'G', 'A')
        self.assertEqual(self.b_flat.root, 'Bb')
        self.assertEqual(self.b_flat.key_signature, ('Bb', 'Eb'))
        self.assertEqual(self.b_flat.notes, bflat_notes)

        d_min_note_names = ('D', 'E', 'F', 'G', 'A', 'Bb', 'C#')
        self.assertEqual(self.d_harm_minor.root, 'D')
        self.assertEqual(self.d_harm_minor.key_signature, ('Bb',))
        self.assertEqual(self.d_harm_minor.quality, 'HARMONIC MINOR')
        self.assertEqual(self.d_harm_minor.notes, d_min_note_names)

    def test_getitem(self):
        self.assertEqual(self.c_scale['submediant'], 'A')

        self.assertEqual(self.d_harm_minor['SUBDOMINANT'], 'G')

    def test_contains(self):
        self.assertTrue('g' in self.c_scale)
        self.assertFalse('D#' in self.c_scale)
        self.assertTrue('Bb' in self.d_harm_minor)
        self.assertFalse('F#' in self.d_harm_minor)

    def test_invalid_degree(self):
        with self.assertRaises(InvalidDegreeError):
            bad_degree = self.c_scale['foobar']

    def test_invalid_tonic(self):
        with self.assertRaises(InvalidTonicError):
            z = Scale('Z MAJOR')

    def test_raises_invalidquality(self):
        with self.assertRaises(InvalidQualityError):
            z = Scale('A Badqual')

    def test_get_relative_minor(self):
        a_minor = self.c_scale.get_relative()
        self.assertEqual(a_minor.notes, ('A', 'B', 'C', 'D', 'E', 'F', 'G'))

    def test_get_relative_major(self):
        f_major = self.d_harm_minor.get_relative()
        self.assertEqual(f_major.notes, ('F', 'G', 'A', 'Bb', 'C', 'D', 'E'))

    def test_get_parallel_minor(self):
        c_minor = self.c_scale.get_parallel()
        self.assertEqual(c_minor.notes, ('C', 'D', 'Eb', 'F', 'G', 'Ab', 'Bb'))

    def test_get_parallel_major(self):
        d_major = self.d_harm_minor.get_parallel()
        self.assertEqual(d_major.notes, ('D', 'E', 'F#', 'G', 'A', 'B', 'C#'))

    def test_get_triad_for_degree(self):
        e_minor = self.c_scale.get_triad_for_degree('mediant')
        to_compare = Chord('E minor')
        self.assertEqual(e_minor, to_compare)

    def test_get_triad_raises(self):
        with self.assertRaises(InvalidDegreeError):
            bad = self.c_scale.get_triad_for_degree('foo')
